/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007 2008 2009 2010 2011 2012 Michael Cornelison
   Source URL: http://kornelix.squarespace.com/fotoxx
   Contact: kornelix2@gmail.com
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#define EX extern                                                          //  enable extern declarations
#include "fotoxx.h"


/**************************************************************************

   Fotoxx image edit - directory (thumbnail gallery display) and navigation 

**************************************************************************/


namespace navi 
{
   #define thumbnail_cachesize 10000                                       //  max thumbnails cached in memory
                                                                           //  supported image file types
   #define imagetypes ".jpeg .jpg .png .tif .tiff .bmp .gif .svg .xpm .tga .ico" 

   #define TEXTWIN GTK_TEXT_WINDOW_TEXT                                    //  GDK window of GTK text view
   #define NEVER GTK_POLICY_NEVER
   #define ALWAYS GTK_POLICY_ALWAYS

   #define     thumbxx 6                                                   //  thumbx array size
   int         thumbx[6] = { 512, 360, 256, 180, 128, 90 };                //  thumbnail image step sizes

   char        *galleryname = 0;                                           //  image directory or file list name
   int         gallerytype = 0;                                            //  1/2/3 = directory/file-list/metadata
   int         nfiles = 0;                                                 //  total file count (incl. subdirks)
   int         nimages = 0;                                                //  image file count
   char        **flist = 0;                                                //  image file list
   char        **mdlist = 0;                                               //  corresp. metadata list

   typedef void ufunc(int Nth, int button);                                //  callback function for clicked image
   ufunc       *userfunc;

   int         xwinW = 1000, xwinH = 700;                                  //  gallery window initial size
   int         xwinX, xwinY;                                               //  gallery window initial position
   int         thumbsize = thumbfilesize;                                  //  thumbnail image <= thumbnail file
   int         thumbW, thumbH;                                             //  gallery window thumbnail cell size
   int         xrows, xcols;                                               //  gallery window thumbnail rows, cols
   int         margin = 5;                                                 //  cell margin from left and top edge
   int         genthumbs = 0;                                              //  count newly generated thumbnails
   int         maxscroll;                                                  //  max. scroll position
   int         topfileposn = 0;                                            //  scroll-to file position (Nth)

   //  private functions
   int    gallery_paint(GtkWidget *, cairo_t *);                           //  gallery window paint function
   int    gallery_paint_metadata(GtkWidget *gDwin, cairo_t *cr);           //  same, for metadata report
   void   draw_text(cairo_t *cr, char *text, int x, int y, int ww);        //  draw text in gallery window
   void   menufuncx(GtkWidget *win, cchar *menu);                          //  function for gallery window buttons
   void   changedirk(GtkWidget *widget, GdkEventButton *event);            //  function for directory change
   void   mouse_xevent(GtkWidget *, GdkEventButton *, void *);             //  gallery window mouse event function
	int    KBxrelease(GtkWidget *, GdkEventKey *, void *);                  //  gallery window key release event
   char * image_navigate(cchar *filez, cchar *action, int Nth = 0);        //  image file list setup and navigate
   int    image_fcomp(cchar *file1, cchar *file2);                         //  file name compare (special sort)
}


/**************************************************************************

   public function to create/update image gallery (thumbnail window)

   Make window scrolling window of thumbnails for a list of files
   Handle window buttons (up row, down page, open file, etc.)
   Call ufunc() when thumbnail image is clicked

   char * gallery(cchar *filez, cchar *action, int Nth, 
                         void ufunc(int Nth, int button))

   filez: image file or directory of image files or file with list of image files

   action:  init:    filez = initial file or directory 
            initF:   filez = file with list of image files to use
            sort:    sort the file list, directories first, ignore case
            insert:  insert filez into file list at position Nth (0 to last+1)
            delete:  delete Nth file in list
            find:    return Nth file (0 base) or null if Nth > last
            paint:   paint gallery window, Nth file in top row
   
   Nth: file to return (action = find) (0 = first file)
        file to scroll-to (action = init, paint) (-1 = no change)

   void ufunc(int Nth, int button):
      - returns Nth of clicked thumbnail (0 to last)
      - returns -1 if gallery window is closed
      - returns -2 if key F1 is pressed (for context help)
      - button is mouse button used on clicked thumbnail
      
   Returned values:
      find: filespec,    others: null
      The returned file belongs to caller and is subject for zfree().

***************************************************************************/


char * gallery(cchar *filez, cchar *action, int Nth, void ufunc(int, int))
{
   using namespace navi;                                                   //  overhauled       v.12.07
   
   zthreadcrash();                                                         //  thread usage not allowed

   if (ufunc) userfunc = ufunc;                                            //  save callback function
   
   if (strstr("init initF",action)) {                                      //  generate new gallery file list
      image_navigate(filez,action);
      Nth = gallery_position(curr_file,Nth);
      if (Nth >= 0) topfileposn = Nth;
      return 0;
   }

   if (strstr("sort insert delete",action)) {                              //  revise gallery file list
      image_navigate(filez,action,Nth);
      if (Nth >= 0) topfileposn = Nth;
      return 0;
   }

   if (strEqu(action,"find"))                                              //  find Nth file in file list
      return image_navigate(0,action,Nth);

   if (strEqu(action,"paint")) {                                           //  paint gallery window
      if (Nth >= 0) topfileposn = Nth;                                     //  new top row file if wanted
      gtk_widget_queue_draw(gDwin);                                        //  draw window
      return 0;
   }
   
   zappcrash("gallery() action: %s",action);                               //  bad call
   return 0;
}


//  private function
//  paint gallery window - draw all thumbnail images that can fit

int navi::gallery_paint(GtkWidget *gDwin, cairo_t *cr)
{
   using namespace navi;

   GdkPixbuf   *pxbT;
   double      x1, y1, x2, y2;
   int         ii, nrows, row, col;
   int         row1, row2;
   int         ww, hh, scrollp;
   int         drwingW, drwingH;
   int         thumx, thumy;
   char        *pp, *fname, p0;
   
   set_gwin_title();                                                       //  main window title = gallery name

   if (! galleryname) return 1;
   gtk_entry_set_text(GTK_ENTRY(gPath),galleryname);                       //  set current directory label  v.12.04

   if (gallerytype == 3) {                                                 //  metadata report
      gallery_paint_metadata(gDwin,cr);
      return 1;
   }
   
   thumbW = thumbsize + 10;                                                //  thumbnail cell size + file name text
   thumbH = thumbsize + 30;

   if (! thumbsize) {
      thumbW = 400;                                                        //  zero, list view
      thumbH = 20;
   }

   xwinW = gtk_widget_get_allocated_width(gSwin);                          //  drawing window size
   xwinH = gtk_widget_get_allocated_height(gSwin);

   xrows = int(0.1 + 1.0 * xwinH / thumbH);                                //  get thumbnail rows and cols that
   xcols = int(0.1 + 1.0 * xwinW / thumbW);                                //    (almost) fit in window 
   if (xrows < 1) xrows = 1;
   if (xcols < 1) xcols = 1;
   nrows = (nfiles+xcols-1) / xcols;                                       //  thumbnail rows, 1 or more
   if (nrows < 1) nrows = 1;

   drwingW = xcols * thumbW + margin + 10;                                 //  layout size for entire gallery
   drwingH = (nrows + 1) * thumbH;                                         //  last row
   if (drwingH < xwinH) drwingH = xwinH;                                   //  at least current size
   
   gtk_widget_get_size_request(gDwin,&ww,&hh);                             //  current size
   if (ww != drwingW || hh != drwingH)
      gtk_widget_set_size_request(gDwin,drwingW,drwingH);                  //  needs to change

   maxscroll = (nrows - xrows) * thumbH;                                   //  max. layout scroll position
   if (maxscroll < 0) maxscroll = 0;

   gtk_adjustment_set_step_increment(gSadj,thumbH);                        //  scrollbar works in row steps
   gtk_adjustment_set_page_increment(gSadj,thumbH * xrows);                //  and in page steps
   
   if (topfileposn >= 0) {                                                 //  new target file position (Nth)
      scrollp = topfileposn / xcols * thumbH;                              //  set scroll position to target file
      if (scrollp > maxscroll) scrollp = maxscroll;
      gtk_adjustment_set_value(gSadj,scrollp);                             //  will cause re-entrance
      if (gtk_adjustment_get_value(gSadj) != scrollp) return 1;            //  scrollp OK in next call
      topfileposn = -1;
   }

   cairo_clip_extents(cr,&x1,&y1,&x2,&y2);                                 //  window region to paint
   row1 = y1 / thumbH;
   row2 = y2 / thumbH;
   
   for (row = row1; row <= row2; row++)                                    //  draw file thumbnails
   {
      for (col = 0; col < xcols; col++)                                    //  draw all columns in row
      {
         ii = row * xcols + col;                                           //  next file
         if (ii >= nfiles) return 1;                                       //  exit 2 nested loops

         p0 = *flist[ii];                                                  //  replace possible ! with /
         *flist[ii] = '/';

         fname = flist[ii];                                                //  filespec
         pp = strrchr(fname,'/');                                          //  get file name only
         if (pp) fname = pp + 1;

         thumx = col * thumbW + margin;                                    //  upper left corner in drawing area
         thumy = row * thumbH + margin;

         draw_text(cr,fname,thumx,thumy,thumbW-5);                         //  draw file name

         if (thumbsize) {                                                  //  zero >> list view
            pxbT = image_thumbnail(flist[ii],thumbsize);                   //  get thumbnail
            if (pxbT) {
               thumy = thumy + 16;                                         //  thumbnail 16 pixels down from text
               gdk_cairo_set_source_pixbuf(cr,pxbT,thumx,thumy);
               cairo_paint(cr);
               g_object_unref(pxbT);
            }
         }

         *flist[ii] = p0;                                                  //  restore !
      }
   }

   return 1;
}


//  private function
//  paint metadata report - draw thumbnail images + metadata

int navi::gallery_paint_metadata(GtkWidget *gDwin, cairo_t *cr)
{
   using namespace navi;

   GdkPixbuf   *pxbT;
   double      x1, y1, x2, y2;
   int         ii, nrows, row, col;
   int         row1, row2;
   int         ww, hh, scrollp;
   int         drwingW, drwingH;
   int         thumx, thumy, textww;
   char        p0;
   
   if (thumbsize < 64) thumbsize = 64;
   thumbW = thumbsize + 10;                                                //  thumbnail layout size
   thumbH = thumbsize + 20;

   xwinW = gtk_widget_get_allocated_width(gSwin);                          //  drawing window size
   xwinH = gtk_widget_get_allocated_height(gSwin);

   xrows = int(0.1 + 1.0 * xwinH / thumbH);                                //  get thumbnail rows fitting in window
   if (xrows < 1) xrows = 1;
   xcols = 1;                                                              //  force cols = 1
   nrows = nfiles;                                                         //  thumbnail rows

   drwingW = xwinW;                                                        //  layout size for entire file list
   if (drwingW < 800) drwingW = 800;
   drwingH = (nrows + 1) * thumbH;                                         //  last row
   if (drwingH < xwinH) drwingH = xwinH;

   gtk_widget_get_size_request(gDwin,&ww,&hh);                             //  current size
   if (ww != drwingW || hh != drwingH)
      gtk_widget_set_size_request(gDwin,drwingW,drwingH);                  //  needs to change

   maxscroll = (nrows - xrows) * thumbH;                                   //  max. layout scroll position
   if (maxscroll < 0) maxscroll = 0;

   gtk_adjustment_set_step_increment(gSadj,thumbH);                        //  scrollbar works in row steps
   gtk_adjustment_set_page_increment(gSadj,thumbH * xrows);                //  and in page steps

   if (topfileposn >= 0) {                                                 //  new target file position (Nth)
      scrollp = topfileposn / xcols * thumbH;                              //  set scroll position to target file
      if (scrollp > maxscroll) scrollp = maxscroll;
      gtk_adjustment_set_value(gSadj,scrollp);                             //  will cause re-entrance
      if (gtk_adjustment_get_value(gSadj) != scrollp) return 1;            //  scrollp OK in next call
      topfileposn = -1;
   }
   
   cairo_clip_extents(cr,&x1,&y1,&x2,&y2);                                 //  window region to paint
   row1 = y1 / thumbH;                                                     //  (< full screen if scrolling)
   row2 = y2 / thumbH;
   
   textww = drwingW - thumbW - 2 * margin;                                 //  space for text right of thumbnail

   for (row = row1; row <= row2; row++)                                    //  draw file thumbnails
   {
      for (col = 0; col < xcols; col++)
      {
         ii = row * xcols + col;                                           //  next file
         if (ii >= nfiles) return 1;                                       //  exit 2 nested loops

         p0 = *flist[ii];                                                  //  replace possible ! with /
         *flist[ii] = '/';

         thumx = col * thumbW + margin;                                    //  upper left corner in window space
         thumy = row * thumbH + margin;
         
         pxbT = image_thumbnail(flist[ii],thumbsize);                      //  get thumbnail
         if (pxbT) {
            gdk_cairo_set_source_pixbuf(cr,pxbT,thumx,thumy);
            cairo_paint(cr);
            g_object_unref(pxbT);
         }

         draw_text(cr,flist[ii],thumbW+margin,thumy,textww);               //  write filespec to right of thumbnail
         
         if (mdlist && mdlist[ii])                                         //  write metadata if present
            draw_text(cr,mdlist[ii],thumbW+margin,thumy+20,textww);
         
         *flist[ii] = p0;                                                  //  restore !
      }
   }

   return 1;
}


//  private function
//  write text for thumbnail limited by width of thumbnail

void navi::draw_text(cairo_t *cr, char *text, int px, int py, int ww)
{
   using namespace navi;

   static PangoFontDescription   *pfont = 0;
   static PangoLayout            *playout = 0;
   
   int            nn;
   static int     thumbsize2 = 0;
   static char    thumbfont[12] = "";
   
   if (thumbsize != thumbsize2) {                                          //  scale file name font to thumbnail size
      thumbsize2 = thumbsize;
      nn = 7 + thumbsize / 128;                                            //  font size from 7 to 10
      if (thumbsize == 0) nn = 9;                                          //  list view, use 9
      sprintf(thumbfont,"sans %d",nn);
      pfont = pango_font_description_from_string(thumbfont);
      playout = pango_cairo_create_layout(cr);                             //  gtk3
      pango_layout_set_font_description(playout,pfont);
   }

   pango_layout_set_width(playout,ww*PANGO_SCALE);                         //  limit width to avail. space  v.12.01
   pango_layout_set_ellipsize(playout,PANGO_ELLIPSIZE_END);
   pango_layout_set_text(playout,text,-1);

   cairo_move_to(cr,px,py);                                                //  gtk3
   cairo_set_source_rgb(cr,0,0,0);
   pango_cairo_show_layout(cr,playout);
   return;
}


//  private function - menu function for gallery window
//    - scroll window as requested
//    - jump to new file or folder as requested

void navi::menufuncx(GtkWidget *win, cchar *menu)
{
   using namespace navi;

   static int  busy = 0;
   int         ii, scrollp, scroll1, scroll2;
   char        *file;
   
   if (strEqu(menu,ZTX("open"))) {                                         //  change directory       v.12.07
      file = zgetfile1(ZTX("change directory"),"folder",curr_dirk);
      if (! file) return;
      gallery(file,"init",0,m_gallery2);
      gallery(0,"paint");
      return;
   }
      
   scrollp = gtk_adjustment_get_value(gSadj);                              //  current scroll position

   if (strEqu(menu,ZTX("bigger")))  {                                      //  next bigger thumbnail size
      for (ii = 0; ii < thumbxx; ii++) 
         if (thumbsize == thumbx[ii]) break;
      if (ii == 0) return;
      thumbsize = thumbx[ii-1];
      topfileposn = scrollp / thumbH * xcols;                              //  keep top row position
      gtk_widget_queue_draw(gDwin);
      return;
   }

   if (strEqu(menu,ZTX("smaller")))  {                                     //  next smaller
      for (ii = 0; ii < thumbxx; ii++) 
         if (thumbsize == thumbx[ii]) break;
      if (ii >= thumbxx-1) thumbsize = 0;                                  //  no thumbs, list view
      else  thumbsize = thumbx[ii+1];
      if (thumbsize < 128 && gallerytype == 3)                             //  min. for metadata report        v.12.01
         thumbsize = 128;
      topfileposn = scrollp / thumbH * xcols;                              //  keep top row position
      gtk_widget_queue_draw(gDwin);
      return;
   }
   
   if (strEqu(menu,ZTX("prev row"))) {                                     //  scroll 1 row in small steps     v.12.07
      if (busy) return;
      busy++;
      scroll1 = scrollp;
      scroll2 = scrollp - thumbH;
      if (scroll2 < 0) scroll2 = 0;
      for (scrollp = scroll1; scrollp > scroll2; scrollp -= 20) {
         topfileposn = -1;
         gtk_adjustment_set_value(gSadj,scrollp);
         gtk_widget_queue_draw(gDwin);
         zmainloop();
      }
      topfileposn = scroll2 / thumbH * xcols;
      gtk_widget_queue_draw(gDwin);
      busy = 0;
      return;
   }
      
   if (strEqu(menu,ZTX("next row"))) {
      if (busy) return;
      busy++;
      scroll1 = scrollp;
      scroll2 = scrollp + thumbH;
      if (scroll2 > maxscroll) scroll2 = maxscroll;
      for (scrollp = scroll1; scrollp < scroll2; scrollp += 20) {
         topfileposn = -1;
         gtk_adjustment_set_value(gSadj,scrollp);
         gtk_widget_queue_draw(gDwin);
         zmainloop();
      }
      topfileposn = scroll2 / thumbH * xcols;
      gtk_widget_queue_draw(gDwin);
      busy = 0;
      return;
   }
      
   if (strEqu(menu,ZTX("prev page"))) scrollp -= thumbH * xrows;
   if (strEqu(menu,ZTX("next page"))) scrollp += thumbH * xrows;
   if (strEqu(menu,ZTX("first page"))) scrollp = 0;
   if (strEqu(menu,ZTX("last page"))) scrollp = maxscroll;

   if (scrollp < 0) scrollp = 0;                                           //  enforce limits
   if (scrollp > maxscroll) scrollp = maxscroll;
   topfileposn = scrollp / thumbH * xcols;
   gtk_widget_queue_draw(gDwin);

   return;
}


//  private function
//  directory selection - get new directory from mouse-click position

void navi::changedirk(GtkWidget *widget, GdkEventButton *event)            //  new v.12.04
{
   using namespace navi;

   int      cc, pos;
   char     *pp, directory[500];
   cchar    *ppc;
   
   if (gallerytype > 1) return;                                            //  collection or search results

   ppc = gtk_entry_get_text(GTK_ENTRY(widget));                            //  current directory 
   if (! ppc) return;
   strncpy0(directory,ppc,500);
   cc = strlen(directory);
   if (cc < 2) return;
   pos = int(event->x) / 8;                                                //  mouse position in directory
   if (pos >= cc) pos = cc - 1;
   if (pos < 2) return;
   pp = strchr(directory+pos,'/');                                         //  truncate following directories
   if (! pp) return;
   *pp = 0;
   image_navigate(directory,"init");                                       //  initialize with that directory
   topfileposn = 0;
   gtk_widget_queue_draw(gDwin);
   return;
}


//  private function
//  mouse event function for gallery window - get selected thumbnail and file
//  user function receives clicked file, which is subject for zfree()

void navi::mouse_xevent(GtkWidget *, GdkEventButton *event, void *)
{
   using namespace navi;

   int            mousex, mousey, mousebutt;
   int            row, col, nrows, ii, err;
   char           *filez;
   struct stat    statb;
   
   if (! nfiles) return;                                                   //  empty window
   
   mousex = int(event->x);
   mousey = int(event->y);
   mousebutt = event->button;

   row = (mousey - margin) / thumbH;                                       //  find selected row, col
   col = (mousex - margin) / thumbW;

   nrows = 1 + (nfiles-1) / xcols;                                         //  total thumbnail rows, 1 or more
   if (col < 0 || col >= xcols) return;
   if (row < 0 || row >= nrows) return;

   ii = xcols * row + col;
   if (ii >= nfiles) return;

   filez = strdupz(flist[ii],0,"navi");                                    //  selected file
   *filez = '/';
   
   err = stat(filez,&statb);
   if (err) {                                                              //  file is gone?
      zfree(filez);
      return;
   }

   if (S_ISDIR(statb.st_mode)) {                                           //  if directory, go there
      image_navigate(filez,"init");
      gtk_widget_queue_draw(gDwin);
      zfree(filez);
      return;
   }
   
   if (userfunc) userfunc(ii,mousebutt);                                   //  clicked file position to user
   return;
}


//  private function
//  KB event function - respond to keyboard navigation keys
//  key definitions: /usr/include/gtk-2.0/gdk/gdkkeysyms.h
//  KBrelease() for main window calls this when tab G is active.

int navi::KBxrelease(GtkWidget *win, GdkEventKey *event, void *)
{
   using namespace navi;

   int      KBkey;
   
   KBkey = event->keyval;

   if (KBkey == GDK_KEY_plus) menufuncx(win,ZTX("bigger"));                //  +/- = bigger/smaller thumbnails
   if (KBkey == GDK_KEY_equal) menufuncx(win,ZTX("bigger"));
   if (KBkey == GDK_KEY_minus) menufuncx(win,ZTX("smaller"));
   if (KBkey == GDK_KEY_KP_Add) menufuncx(win,ZTX("bigger"));              //  keypad +/- also 
   if (KBkey == GDK_KEY_KP_Subtract) menufuncx(win,ZTX("smaller"));

   if (KBkey == GDK_KEY_Left) menufuncx(win,ZTX("prev page"));             //  left arrow = previous page
   if (KBkey == GDK_KEY_Right) menufuncx(win,ZTX("next page"));            //  right arrow = next page
   if (KBkey == GDK_KEY_Up) menufuncx(win,ZTX("prev row"));                //  up arrow = previous row
   if (KBkey == GDK_KEY_Down) menufuncx(win,ZTX("next row"));              //  down arrow = next row
   
   if (KBkey == GDK_KEY_Home) menufuncx(win,ZTX("first page"));            //  keys added 
   if (KBkey == GDK_KEY_End) menufuncx(win,ZTX("last page"));
   if (KBkey == GDK_KEY_Page_Up) menufuncx(win,ZTX("prev page"));
   if (KBkey == GDK_KEY_Page_Down) menufuncx(win,ZTX("next page"));

   if (KBkey == GDK_KEY_F1)
      showz_userguide(zfuncs::F1_help_topic);

   if (KBkey == GDK_KEY_F11) {                                             //  F11  >>  toggle main window     v.12.06
      if (! Wmaximized) win_maximize();                                    //    to full-screen mode and back
      else win_unmaximize();
   }

   return 1;
}


//  set the main window title for the gallery window                       //  v.12.07
//  window title = gallery name

void set_gwin_title()
{
   char     title[200];

   if (cTab != tabG) return;
   if (! navi::galleryname) strcpy(title,"no current directory");
   else snprintf(title,199,"%s  %d files",navi::galleryname,navi::nfiles);
   gtk_window_set_title(MWIN,title);
   return;
}


//  Get file position in file list.
//  If Nth position matches file, this is returned.
//  Otherwise the file is searched from position 0.
//  Position 0-last is returned if found, or -1 if not.

int gallery_position(cchar *file, int Nth)
{
   using namespace navi;
   
   int      ii;
   
   if (! nfiles) return -1;
   if (! file) return -1;                                                  //  bugfix     v.12.07

   if (Nth >= 0 && Nth < nfiles) ii = Nth;
   else ii = 0;
   
   if (strEqu(file+1,flist[ii]+1)) return ii;                              //  1st flist[ii] char. may be !

   for (ii = 0; ii < nfiles; ii++)
      if (strEqu(file+1,flist[ii]+1)) break;

   if (ii < nfiles) return ii;
   return -1;
}


//  Determine if a file is a directory or a supported image file type
//  Return: 0 = error, 1 = directory, 2 = image file, 3 = other

int image_file_type(cchar *file)
{
   using namespace navi;

   int            err, cc;
   cchar          *pp;
   struct stat    statbuf;

   if (! file) return 0;
   err = stat(file,&statbuf);
   if (err) return 0;
   
   if (thumbdirk && strNeq(thumbdirk,"undefined")) {
      cc = strlen(thumbdirk);                                              //  outlaw files in thumbnail directory
      if (strnEqu(file,thumbdirk,cc)) return 3;                            //  v.12.06
   }

   if (strstr(file,"/.thumbnails")) return 3;                              //  outlaw files in /.thumbnails    v.12.06

   if (S_ISDIR(statbuf.st_mode)) return 1;                                 //  directory

   if (S_ISREG(statbuf.st_mode)) {                                         //  reg. file
      pp = strrchr(file,'.');
      if (! pp) return 3;
      pp = strcasestr(imagetypes,pp);                                      //  supported image type
      if (pp) return 2;
   }
   
   return 3;
}


//  Given a thumbnail file, get the corresponding image file.
//  Returns null if no image file found.
//  Returned file is subject for zfree().
//  image file:  /image/dirk/file.xxx                                      //  .jpg .png .tiff etc.
//  thumb dirk:  /thumb/dirk                                               //  if thumb dirk defined
//  thumb file:  /thumb/dirk/image/dirk/file.xxx.jpeg                      //     thumb file is like this
//  thumb file:  /image/dirk/.thumbnails/file.xxx.jpeg                     //  else like this

char * thumb2imagefile(cchar *thumbfile)
{
   struct stat statb;
   int         err, cc;
   char        *imagefile;
   char        *pp;
   
   if (thumbdirk && strNeq(thumbdirk,"undefined"))                         //  remove /thumb/dirk from the front
   {
      cc = strlen(thumbdirk);
      imagefile = strdupz(thumbfile+cc,0,"thumb2imagefile");               //  have /image/dirk/file.xxx.jpeg
      cc = strlen(imagefile);
      imagefile[cc-5] = 0;                                                 //  /image/dirk/file.xxx
   }

   else                                                                    //  remove /.thumbnails from between
   {
      pp = (char *) strrchr(thumbfile,'/');                                //  /image/dirk/.thumbnails/file.xxx.jpeg
      if (! pp) return 0;                                                  //             |           |             |
      pp = pp - 12;                                                        //             pp          +12           +cc
      if (! strnEqu(pp,"/.thumbnails/",13)) return 0;
      cc = strlen(pp+12);
      if (! strEqu(pp+12+cc-5,".jpeg")) return 0;
      imagefile = strdupz(thumbfile,0,"thumb2imagefile");
      pp = imagefile + (pp - thumbfile);
      memmove((char *) pp,pp+12,cc-5);                                              //  /image/dirk/file.xxx
      pp[cc-5] = 0;
   }

   err = stat(imagefile,&statb);                                           //  check file exists
   if (! err) return imagefile;                                            //  return image file
   zfree(imagefile);                                                       //  not found
   return 0;
}


//  Given an image file, get the corresponding thumbnail file.
//  The filespec is returned whether or not the file exists.
//  Returned file is subject for zfree().

char * image2thumbfile(cchar *imagefile)
{
   int      cc, cc1, cc2;
   char     *pp, *thumbfile;

   if (thumbdirk && strNeq(thumbdirk,"undefined"))
   {
      cc1 = strlen(thumbdirk);
      cc2 = strlen(imagefile);
      thumbfile = zmalloc(cc1+cc2+6,"image2thumbfile");
      strcpy(thumbfile,thumbdirk);                                         //  /thumb/dirk
      strcpy(thumbfile+cc1,imagefile);                                     //  /thumb/dirk/image/dirk/file.xxx
      strcpy(thumbfile+cc1+cc2,".jpeg");                                   //  /thumb/dirk/image/dirk/file.xxx.jpeg
      return thumbfile;
   }

   else
   {
      thumbfile = strdupz(imagefile,18,"image2thumbfile");                 //  /image/dirk/file.xxx
      pp = strrchr(thumbfile,'/');                                         //             |        |
      if (! pp) return 0;                                                  //             pp       +cc
      cc = strlen(pp);
      memmove(pp+12,pp,cc);                                                //  /image/dirk............/file.xxx
      strncpy(pp,"/.thumbnails",12);                                       //  /image/dirk/.thumbnails/file.xxx
      strcpy(pp+12+cc,".jpeg");                                            //  /image/dirk/.thumbnails/file.xxx.jpeg
      return thumbfile;
   }
}


//  Get thumbnail file for the given image file.
//  If missing or stale, add or update thumbnail file on disk.
//  Returned filespec is subject for zfree().

char * image_thumbfile(char *imagefile)
{
   GdkPixbuf         *thumbpxb;
   GError            *gerror = 0;
   char              *thumbfile, *pp;
   int               err, ww, hh;
   struct stat       statf, statb;
   PXM               *pxm1, *pxm2;
   float             scale;
   static int        Fmkdirerr = 0;
   
   err = stat(imagefile,&statf);
   if (err) return 0;
   if (image_file_type(imagefile) != 2) return 0;                          //  not a supported image type
   
   thumbfile = image2thumbfile(imagefile);                                 //  get thumbnail file for image file
   if (! thumbfile) return 0;

   err = stat(thumbfile,&statb);                                           //  thumbfile exists, up to date ??
   if (! err && statb.st_mtime >= statf.st_mtime) return thumbfile;        //  yes, return it

   pp = strrchr(thumbfile,'/');                                            //  check if thumbnail directory exists
   *pp = 0;
   err = stat(thumbfile,&statb);
   *pp = '/';
   if (err) {                                                              //  no
      pp = thumbfile;
      while (true) {
         pp = strchr(pp+1,'/');                                            //  check each directory level
         if (! pp) break;
         *pp = 0;
         err = stat(thumbfile,&statb);
         if (err) {
            err = mkdir(thumbfile,0751);                                   //  if missing, try to create
            if (err) {
               if (! Fmkdirerr++)                                          //  if unable, print error once only
                  printf("create thumbnail: %s \n",strerror(errno));
               zfree(thumbfile);
               return 0;
            }
         }
         *pp = '/';
      }
   }

   ww = hh = thumbfilesize;                                                //  create thumbnail pixbuf
   thumbpxb = gdk_pixbuf_new_from_file_at_size(imagefile,ww,hh,&gerror);
   if (thumbpxb) {                                                         //  save to thumbnail file
      gdk_pixbuf_save(thumbpxb,thumbfile,"jpeg",&gerror,"quality","80",null);
      g_object_unref(thumbpxb);
      navi::genthumbs++;                                                   //  count generated thumbnails
      return thumbfile;
   }
                                                                           //  backup thumbnail creator     v.12.03
   pxm1 = f_load(imagefile);                                               //  read image file
   if (! pxm1) return 0;                                                   //  should not happen
   ww = pxm1->ww;                                                          //  scale image to thumbnail size
   hh = pxm1->hh;
   if (ww > hh) scale = 1.0 * thumbfilesize / ww;
   else scale = 1.0 * thumbfilesize / hh;
   if (scale < 1) {
      ww = scale * ww;
      hh = scale * hh;
      pxm2 = PXM_rescale(pxm1,ww,hh);
      PXM_free(pxm1);
      pxm1 = pxm2;
   }
   PXBwrite(pxm1,thumbfile);                                               //  write thumbnail as .jpeg file
   PXM_free(pxm1);
   navi::genthumbs++;
   return thumbfile;
}


//  Get thumbnail image (pixbuf) for given image file.
//  Use pixbuf cached in memory if available.
//  Create pixbuf from thumbnail file if size <= thumbnail file size.
//  Create pixbuf from image file if size > thumbnail file size.
//  Add thumbnail file if missing, or update it if older than image file.
//  Add pixbuf to memory cache if not there already.
//  Returned thumbnail belongs to caller: g_object_unref() is necessary.

GdkPixbuf * image_thumbnail(char *fpath, int size)                         //  v.12.06
{
   GdkPixbuf         *thumbpxb;
   GError            *gerror = 0;
   int               ii, err;
   char              *bpath;
   time_t            mtime;
   struct stat       statf;
   const int         cachesize = thumbnail_cachesize;                      //  shorthand

   static int           nextcache, ftf = 1;
   static int           sizecache[cachesize];
   static time_t        mtimecache[cachesize];
   static char          *fpathcache[cachesize];
   static GdkPixbuf     *pixbufcache[cachesize];
   
   zthreadcrash();                                                         //  thread usage not allowed

   if (ftf) {                                                              //  first call
      for (ii = 0; ii < cachesize; ii++) 
         fpathcache[ii] = 0;                                               //  clear cache
      ftf = 0;
   }
   
   err = stat(fpath,&statf);                                               //  fpath status info
   if (err) return 0;

   if (! size) size = thumbfilesize;                                       //  default thumb size

   if (S_ISDIR(statf.st_mode)) {                                           //  if directory, return folder image
      bpath = zmalloc(300);
      strncatv(bpath,299,zfuncs::zicondir,"/folder256.png",null);
      thumbpxb = gdk_pixbuf_new_from_file_at_size(bpath,size,size,&gerror);
      zfree(bpath);
      return thumbpxb;
   }

   mtime = statf.st_mtime;                                                 //  last modification time

   for (ii = nextcache; ii >= 0; ii--)
      if (fpathcache[ii] && strEqu(fpath,fpathcache[ii]) &&
          sizecache[ii] == size && mtime == mtimecache[ii]) break;         //  check mtime
   if (ii >= 0) {
      thumbpxb = gdk_pixbuf_copy(pixbufcache[ii]);
      return thumbpxb;
   }
   for (ii = cachesize-1; ii > nextcache; ii--)                            //  continue search
      if (fpathcache[ii] && strEqu(fpath,fpathcache[ii]) &&
          sizecache[ii] == size && mtime == mtimecache[ii]) break;
   if (ii > nextcache) {
      thumbpxb = gdk_pixbuf_copy(pixbufcache[ii]);
      return thumbpxb;
   }

   if (size > thumbfilesize || ! Fthumbnails)                              //  make thumbnail from image file     v.12.06
      thumbpxb = gdk_pixbuf_new_from_file_at_size(fpath,size,size,&gerror);
   else {
      bpath = image_thumbfile(fpath);                                      //  make thumbnail from thumbnail file
      if (! bpath) return 0;
      thumbpxb = gdk_pixbuf_new_from_file_at_size(bpath,size,size,&gerror);
      zfree(bpath);
   }
   if (! thumbpxb) return 0;

   nextcache++;                                                            //  next cache slot (oldest)
   if (nextcache == cachesize) nextcache = 0;
   ii = nextcache;
   if (fpathcache[ii]) {                                                   //  free prior occupant
      zfree(fpathcache[ii]);
      g_object_unref(pixbufcache[ii]);
   }
   fpathcache[ii] = strdupz(fpath,0,"thumbnail_cache");                    //  add new occupant
   pixbufcache[ii] = gdk_pixbuf_copy(thumbpxb);                            //  this memory is not tracked
   sizecache[ii] = size;
   mtimecache[ii] = mtime;

   return thumbpxb;                                                        //  return pixbuf to caller
}


/**************************************************************************

   private function - manage list of image files within a directory

   get an image file in the same directory as given file or directory

   char * navi::image_navigate(cchar *filez, cchar *action, int Nth)

   action:  init:    file list = directories and image files in directory filez
            initF:   file list = filez = list of image files to use (cchar **)
            sort:    sort the file list, directories first, ignore case
            insert:  insert filez into file list at position Nth (0 to last+1)
            delete:  delete Nth file in list
            find:    returns Nth file (0 base) or null if Nth > last

   Nth: file to return for action = find
   
   Returned values:
      find: filespec, else null
      The returned file belongs to caller and is subject for zfree().

***************************************************************************/

char * navi::image_navigate(cchar *filez, cchar *action, int Nth)
{
   using namespace navi;

   char           *buff;
   cchar          *findcommand = "find \"%s\" -maxdepth 1";
   char           *pp, *file2;
   int            err, ii, cc, ftyp, contx = 0, fposn;
   FILE           *fid;
   struct stat    statbuf;
   
   if (! strstr("init initF sort insert delete find",action))
         zappcrash("image_navigate %s",action);

   if (strnEqu(action,"init",4))                                           //  init or initF
   {
      if (flist) {
         for (ii = 0; ii < nfiles; ii++) zfree(flist[ii]);                 //  free prior file list
         zfree(flist);
      }

      if (mdlist) {                                                        //  clear prior metadata if any
         for (ii = 0; ii < nfiles; ii++) 
            if (mdlist[ii]) zfree(mdlist[ii]);
         zfree(mdlist);
         mdlist = 0;
      }

      cc = MAXIMAGES * sizeof(char *);      
      flist = (char **) zmalloc(cc,"navi");                                //  list of file pointers

      nfiles = nimages = 0;                                                //  no files
      fposn = 0;
      gallerytype = 0;                                                     //  v.12.01
   }

   if (strEqu(action,"init"))                                              //  initialize from given directory
   {
      if (! filez) return 0;                                               //  should not happen

      if (galleryname) zfree(galleryname);
      galleryname = strdupz(filez,0,"navi");
      gallerytype = 1;                                                     //  gallery type = directory

      err = stat(galleryname,&statbuf);
      if (err) {
         pp = (char *) strrchr(galleryname,'/');                           //  bad file, check directory part  
         if (! pp) return 0;
         pp[1] = 0;
         err = stat(galleryname,&statbuf);
         if (err) return 0;                                                //  give up, empty file list
      }

      if (S_ISREG(statbuf.st_mode)) {                                      //  if a file, get directory part
         pp = (char *) strrchr(galleryname,'/');
         if (! pp) return 0;
         pp[1] = 0;
      }

      while ((buff = command_output(contx,findcommand,galleryname)))       //  find all files
      {
         if (strEqu(buff,galleryname)) {                                   //  skip self directory
            zfree(buff);
            continue;
         }
         
         if (nfiles == MAXIMAGES) {                                        //  message, not crash 
            zmessageACK(0,Btoomanyfiles,MAXIMAGES);
            break;
         }

         ftyp = image_file_type(buff);

         if (ftyp == 1) {                                                  //  subdirectory
            flist[nfiles] = buff;                                          //  add to file list
            flist[nfiles][0] = '!';                                        //  if directory, make it sort first
            nfiles++;
         }

         else if (ftyp == 2) {                                             //  supported image file type
            flist[nfiles] = buff;                                          //  add to file list
            nfiles++;
            nimages++; 
         }

         else {
            zfree(buff);                                                   //  (thumbnails not ftyp 1)
            continue;
         }
      }

      if (nfiles > 1) 
         HeapSort(flist,nfiles,image_fcomp);                               //  Heap Sort - pointers to strings

      return 0;
   }

   if (strEqu(action,"initF"))                                             //  initialize from given list
   {
      if (galleryname) zfree(galleryname);
      galleryname = strdupz(filez,0,"navi");
      gallerytype = 2;                                                     //  gallery type = file list
      
      fid = fopen(galleryname,"r");                                        //  open file
      if (! fid) return 0;

      buff = zmalloc(maxfcc);
      
      while (true)                                                         //  read list of files
      {
         if (nfiles == MAXIMAGES) {
            zmessageACK(0,Btoomanyfiles,MAXIMAGES);
            break;
         }
         pp = fgets_trim(buff,maxfcc-1,fid,1);
         if (! pp) break;
         err = stat(pp,&statbuf);                                          //  check file exists
         if (err) continue;
         flist[nfiles] = strdupz(buff,0,"navi");                           //  add files to memory list
         nfiles++;
         nimages++;
      }

      fclose(fid);
      zfree(buff);                                                         //  sort removed
      return 0;
   }
   
   if (strEqu(action,"sort"))                                              //  sort the list from initF 
   {
      if (gallerytype == 3) return 0;                                      //  v.12.01
      if (nfiles < 2) return 0;
      HeapSort(flist,nfiles,image_fcomp);                                  //  Heap Sort - pointers to strings
      return 0;
   }

   if (strEqu(action,"insert"))                                            //  insert new file into list
   {
      if (gallerytype == 3) return 0;                                      //  metadata report     v.12.01
      fposn = Nth;                                                         //  file position from caller
      if (fposn < 0) fposn = 0;                                            //  limit to allowed range
      if (fposn > nfiles) fposn = nfiles;

      if (nfiles == MAXIMAGES-1) {                                         //  no room
         zmessageACK(0,Btoomanyfiles,MAXIMAGES);
         return 0;
      }

      for (ii = nfiles; ii > fposn; ii--)                                  //  create hole is list
         flist[ii] = flist[ii-1];

      flist[fposn] = strdupz(filez,0,"navi");                              //  put new file in hole
      nfiles++;
      nimages++;                                                           //  bugfix
   }

   if (strEqu(action,"delete"))                                            //  delete file from list
   {
      if (gallerytype == 3) return 0;                                      //  metadata report     v.12.01
      fposn = Nth;                                                         //  file position from caller must be OK
      if (fposn < 0 || fposn > nfiles-1) return 0;
      nfiles--;
      if (*flist[fposn] != '!') nimages--;                                 //  not a directory              v.12.01
      zfree(flist[fposn]);                                                 //  remove file from list
      for (ii = fposn; ii < nfiles; ii++) {                                //  close the hole
         if (nfiles < 0) printf("meaningless reference %d",ii);            //  stop g++ optimization bug    ///
         flist[ii] = flist[ii+1];
      }
   }
   
   if (strEqu(action,"find"))
   {
      fposn = Nth;                                                         //  file position from caller must be OK
      if (fposn < 0 || fposn > nfiles-1) return 0;
      file2 = strdupz(flist[fposn],0,"navi");                              //  get Nth file
      file2[0] = '/';                                                      //  restore initial '/'
      err = stat(file2,&statbuf);
      if (! err) return file2;
      zfree(file2);
   }
   
   return 0;
}


//  private function for special file name compare
//  directories sort first and upper/lower case is ignored

int navi::image_fcomp(cchar *file1, cchar *file2)
{
   int      nn;
   nn = strcasecmp(file1,file2);                                           //  compare ignoring case
   if (nn != 0) return nn;
   nn = strcmp(file1,file2);                                               //  if equal, do normal compare
   return nn;
}


//  find function which can be called from a thread

char * gallery_find(int Nth)
{
   using namespace navi;

   char           *file2;
   int            err, fposn;
   struct stat    statbuf;
   
   fposn = Nth;                                                            //  file position from caller must be OK
   if (fposn < 0 || fposn > nfiles-1) return 0;
   file2 = strdupz(flist[fposn],0,"navi");                                 //  get Nth file
   file2[0] = '/';                                                         //  restore initial '/'
   err = stat(file2,&statbuf);
   if (! err) return file2;
   zfree(file2);
   return 0;
}


/**************************************************************************/

//  Select files from the image gallery window, return list of files selected.
//  The dialog shows the list of files selected and can be edited.
//  The returned file list belongs to caller and is subject for zfree().
//  The file list EOL is marked with null.
//  The gallery() callback function is restored to caller's function.         v.12.07

int  imgl_showthumb();
void imgl_insert_file(cchar *imagefile);
void imgl_gallery_file(int Nth, int button);

GtkWidget      *imgl_drawarea = 0;
GtkWidget      *imgl_files = 0;
cchar          *imgl_font = "Monospace 8";
zdialog        *imgl_zd = 0;
int            imgl_fontheight = 14;
int            imgl_cursorpos = 0;
navi::ufunc    *prior_ufunc;


char ** gallery_getfiles(char *startfile, GtkWidget *parent)
{
   using namespace navi;

   int  imgl_dialog_event(zdialog *zd, cchar *event);
   int  imgl_mouseclick(GtkWidget *, GdkEventButton *event, void *);

   PangoLanguage           *plang;
   PangoFontDescription    *pfontdesc;
   PangoContext            *pcontext;
   PangoFont               *pfont;
   PangoFontMetrics        *pmetrics;
   GdkCursor               *cursor;
   GdkWindow               *gdkwin;
   GtkTextBuffer           *textBuff;
   GtkTextIter             iter1, iter2;

   int               fontascent, fontdescent;
   int               line, nlines, ii, Nth;
   char              *imagefile, **filelist;
   
   zthreadcrash();                                                         //  thread usage not allowed
   
   if (! startfile) {                                                      //  v.12.05
      startfile = topdirk[0];
      gallery(startfile,"init");
      Nth = 0;
   }
   else Nth = curr_file_posn;                                              //  v.12.05
   
   prior_ufunc = navi::userfunc;                                           //  save gallery callback func.     v.12.07

   gallery(startfile,"paint",Nth,imgl_gallery_file);                       //  activate image gallery window

   imgl_zd = zdialog_new(ZTX("Select Files"),parent,ZTX("done"),ZTX("cancel"),null);
   zdialog_add_widget(imgl_zd,"hbox","hb1","dialog",0,"expand|space=5");
   zdialog_add_widget(imgl_zd,"frame","fr11","hb1",0,"expand");
   zdialog_add_widget(imgl_zd,"scrwin","scrwin","fr11",0,"expand");
   zdialog_add_widget(imgl_zd,"edit","files","scrwin");
   zdialog_add_widget(imgl_zd,"vbox","vb12","hb1");
   zdialog_add_widget(imgl_zd,"frame","fr12","vb12");
   zdialog_add_widget(imgl_zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(imgl_zd,"button","delete","hb2",ZTX("delete"),"space=8");
   zdialog_add_widget(imgl_zd,"button","insert","hb2",ZTX("insert"),"space=8");
   zdialog_add_widget(imgl_zd,"button","addall","hb2",ZTX("add all"),"space=30");

   GtkWidget *textbox = zdialog_widget(imgl_zd,"files");                   //  disable text wrap 
   gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(textbox),GTK_WRAP_NONE);

   GtkWidget *frame = zdialog_widget(imgl_zd,"fr12");                      //  drawing area for thumbnail image
   imgl_drawarea = gtk_drawing_area_new();
   gtk_widget_set_size_request(imgl_drawarea,256,258);                     //  increased
   gdkwin = gtk_widget_get_window(imgl_drawarea);                          //  gtk3
   gtk_container_add(GTK_CONTAINER(frame),imgl_drawarea);
   
   imgl_files = zdialog_widget(imgl_zd,"files");                           //  activate mouse-clicks for
   gtk_widget_add_events(imgl_files,GDK_BUTTON_PRESS_MASK);                //    file list widget
   G_SIGNAL(imgl_files,"button-press-event",imgl_mouseclick,0);

   pfontdesc = pango_font_description_from_string(imgl_font);              //  set default font for files window
   gtk_widget_modify_font(imgl_files,pfontdesc);

   plang = pango_language_get_default();                                   //  get font metrics (what a mess)
   pcontext = gtk_widget_get_pango_context(imgl_files);
   pfont = pango_context_load_font(pcontext,pfontdesc);
   pmetrics = pango_font_get_metrics(pfont,plang);
   fontascent = pango_font_metrics_get_ascent(pmetrics) / PANGO_SCALE;
   fontdescent = pango_font_metrics_get_descent(pmetrics) / PANGO_SCALE;
   imgl_fontheight = fontascent + fontdescent;                             //  effective line height

   zdialog_resize(imgl_zd,600,0);                                          //  start dialog
   zdialog_run(imgl_zd,imgl_dialog_event);
   settab("G");                                                            //  set G tab

   cursor = gdk_cursor_new(GDK_TOP_LEFT_ARROW);                            //  arrow cursor for file list widget
   gdkwin = gtk_text_view_get_window(GTK_TEXT_VIEW(imgl_files),TEXTWIN);   //  (must be done after window realized)
   gdk_window_set_cursor(gdkwin,cursor);
   imgl_cursorpos = 0;

   zdialog_wait(imgl_zd);                                                  //  wait for completion

   gallery(0,"paint",-1,prior_ufunc);                                      //  restore gallery callback func.  v.12.07

   if (imgl_zd->zstat != 1) {                                              //  cancelled
      zdialog_free(imgl_zd);                                               //  kill dialog
      return 0;
   }

   textBuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(imgl_files));
   nlines = gtk_text_buffer_get_line_count(textBuff);
   
   filelist = (char **) zmalloc((nlines+1) * sizeof(char *),"imgl.getfiles");

   for (ii = line = 0; line < nlines; line++)                              //  get list of files from dialog
   {
      gtk_text_buffer_get_iter_at_line(textBuff,&iter1,line);              //  iter at line start
      iter2 = iter1;
      gtk_text_iter_forward_to_line_end(&iter2);
      imagefile = gtk_text_buffer_get_text(textBuff,&iter1,&iter2,0);      //  get line of text
      if (imagefile && *imagefile == '/') {
         filelist[ii] = strdupz(imagefile,0,"imgl.getfiles");              //  >> next file in list
         free(imagefile);
         ii++;
      }
   }
   filelist[ii] = 0;                                                       //  mark EOL
   
   zdialog_free(imgl_zd);                                                  //  kill dialog

   if (! ii) {
      zfree(filelist);                                                     //  file list is empty
      return 0;
   }

   return filelist;                                                        //  return file list
}


//  imgl dialog event function

int imgl_dialog_event(zdialog *zd, cchar *event)
{
   GtkTextBuffer  *textBuff;
   GtkTextIter    iter1, iter2;
   char           *ftemp;
   static char    *imagefile = 0;
   int            line, posn;
   
   if (strEqu(event,"delete"))                                             //  delete file at cursor position
   {
      if (imagefile) zfree(imagefile);
      imagefile = 0;

      textBuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(imgl_files));
      line = imgl_cursorpos;
      gtk_text_buffer_get_iter_at_line(textBuff,&iter1,line);              //  iter at line start
      iter2 = iter1;
      gtk_text_iter_forward_to_line_end(&iter2);                           //  iter at line end

      ftemp = gtk_text_buffer_get_text(textBuff,&iter1,&iter2,0);          //  get selected file
      if (! ftemp || *ftemp != '/') {
         if (ftemp) free(ftemp);
         return 0;
      }

      imagefile = strdupz(ftemp,0,"imgl.getfiles");                        //  save deleted file for poss. insert
      free(ftemp);
      
      gtk_text_buffer_delete(textBuff,&iter1,&iter2);                      //  delete file text
      gtk_text_buffer_get_iter_at_line(textBuff,&iter2,line+1);
      gtk_text_buffer_delete(textBuff,&iter1,&iter2);                      //  delete empty line (\n)

      imgl_showthumb();                                                    //  thumbnail = next file
   }

   if (strEqu(event,"insert"))                                             //  insert last deleted file
   {                                                                       //    at current cursor position
      if (! imagefile) return 0;
      imgl_insert_file(imagefile);
      zfree(imagefile);
      imagefile = 0;
   }

   if (strEqu(event,"addall"))                                             //  insert all files in image gallery
   {
      textBuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(imgl_files));
      posn = 0;

      while (true)
      {
         imagefile = gallery(0,"find",posn);                               //  get first or next file
         if (! imagefile) break;
         posn++;
         gtk_text_buffer_get_iter_at_line(textBuff,&iter1,imgl_cursorpos);
         gtk_text_buffer_insert(textBuff,&iter1,"\n",1);                   //  insert new blank line
         gtk_text_buffer_get_iter_at_line(textBuff,&iter1,imgl_cursorpos);
         gtk_text_buffer_insert(textBuff,&iter1,imagefile,-1);             //  insert image file
         zfree(imagefile);
         imgl_cursorpos++;                                                 //  advance cursor position
      }
   }

   return 0;
}


//  add image file to list at current cursor position, set thumbnail = file

void imgl_insert_file(cchar *imagefile)
{
   GtkTextIter    iter;
   GtkTextBuffer  *textBuff;
   
   textBuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(imgl_files));
   gtk_text_buffer_get_iter_at_line(textBuff,&iter,imgl_cursorpos);
   gtk_text_buffer_insert(textBuff,&iter,"\n",1);                          //  insert new blank line
   gtk_text_buffer_get_iter_at_line(textBuff,&iter,imgl_cursorpos);
   gtk_text_buffer_insert(textBuff,&iter,imagefile,-1);                    //  insert image file

   imgl_showthumb();                                                       //  update thumbnail
   imgl_cursorpos++;                                                       //  advance cursor position

   return;
}


//  called from image gallery window when a thumbnail is clicked
//  add image file to list at current cursor position, set thumbnail = file

void imgl_gallery_file(int Nth, int button)
{
   int            ftyp;
   char           *imagefile;
   
   if (Nth == -2) {
      showz_userguide(zfuncs::F1_help_topic);                              //  F1 context help
      return;
   }

   imagefile = gallery(0,"find",Nth);                                      //  get file at clicked position
   if (! imagefile) return;

   ftyp = image_file_type(imagefile);                                      //  ignore directories
   if (ftyp == 2) imgl_insert_file(imagefile);                             //  insert file at current position
   zfree(imagefile);

   return;
}


//  process mouse click in files window: 
//  set new cursor position and set thumbnail = clicked file

int imgl_mouseclick(GtkWidget *, GdkEventButton *event, void *)
{
   int            mpy;
   GtkWidget      *scrollwin;
   GtkAdjustment  *scrolladj;
   double         scrollpos;

   if (event->type != GDK_BUTTON_PRESS) return 0;
   mpy = int(event->y);
   scrollwin = zdialog_widget(imgl_zd,"scrwin");                           //  window scroll position
   scrolladj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(scrollwin));
   scrollpos = gtk_adjustment_get_value(scrolladj);
   imgl_cursorpos = (mpy + scrollpos) / imgl_fontheight;                   //  line selected
   imgl_showthumb();                                                       //  show thumbnail image
   return 0;
}


//  show thumbnail for file at current cursor position

int imgl_showthumb()
{
   int            line;
   char           *imagefile;
   GdkWindow      *gdkwin;
   cairo_t        *cr;
   GtkTextBuffer  *textBuff;
   GtkTextIter    iter1, iter2;
   GdkPixbuf      *thumbnail = 0;

   gdkwin = gtk_widget_get_window(imgl_drawarea);                          //  gtk3
   cr = gdk_cairo_create(gdkwin);
   
   line = imgl_cursorpos;
   textBuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(imgl_files));
   gtk_text_buffer_get_iter_at_line(textBuff,&iter1,line);                 //  iter at line start
   iter2 = iter1;
   gtk_text_iter_forward_to_line_end(&iter2);                              //  iter at line end

   imagefile = gtk_text_buffer_get_text(textBuff,&iter1,&iter2,0);         //  get selected file
   if (*imagefile != '/') {
      free(imagefile);
      return 0;
   }

   thumbnail = image_thumbnail(imagefile,256);                             //  get thumbnail
   free(imagefile);

   if (thumbnail) {
      cairo_set_source_rgb(cr,1,1,1);                                      //  white background    gtk3
      cairo_paint(cr); 
      gdk_cairo_set_source_pixbuf(cr,thumbnail,0,0);                       //  paint thumbnail     gtk3
      cairo_paint(cr);
      g_object_unref(thumbnail);
   }

   return 0;
}


